try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)

from .bg_subtracter import (
    BackgroundSubtracter,
    da_to_zarr,
    get_shrink_factor,
    compute_chunk_size_and_overlap
)


import math

import imagej
import jpype
import numpy as np
import psutil
import scyjava

# `BackgroundSubtracter` is in ImageJ not ImageJ2 - use global variable to
# prevent multiple instances

# https://github.com/imagej/pyimagej/blob/fd7de5c192e712ec96fab0ffc0ee84b9ffbf1f12/doc/11-Working-with-the-original-ImageJ.ipynb
'''
ImageJ was not designed to run multiple simultaneous instances in the same JVM,
whereas ImageJ2 supports multiple gateways at once.

def legacy_status(gateway):
    print(f" legacy service: {gateway.legacy}") print(f"  legacy active?
    {gateway.legacy and gateway.legacy.isActive()}") print(f"ImageJ2 version:
    {gateway.getVersion()}")

another_ij = imagej.init() print("[ij - the original gateway]")
legacy_status(ij) print("\n[another_ij - a second gateway we constructed just
now]") legacy_status(another_ij)

[ij - the original gateway]
 legacy service: net.imagej.legacy.LegacyService [priority = 1.0]
  legacy active? True
ImageJ2 version: 2.14.0/1.54f

[another_ij - a second gateway we constructed just now]
 legacy service: net.imagej.legacy.LegacyService [priority = 1.0]
  legacy active? False
ImageJ2 version: 2.14.0/Inactive
'''
IJ = None


class BackgroundSubtracter():
    '''
    https://imagej.net/ij/developer/api/ij/ij/plugin/filter/BackgroundSubtracter.html#rollingBallBackground(ij.process.ImageProcessor,double,boolean,boolean,boolean,boolean,boolean)

    public void rollingBallBackground(ImageProcessor ip, double radius, boolean createBackground, boolean lightBackground, boolean useParaboloid, boolean doPresmooth, boolean correctCorners)

    Create or subtract a background, works for all image types. For RGB images, the background is subtracted from each channel separately
    
    Parameters:
    
    ip - The image. On output, it will become the background-subtracted image or
    the background (depending on createBackground).
    
    radius - Radius of the rolling ball creating the background (actually a
    paraboloid of rotation with the same curvature)
    
    createBackground - Whether to create a background, not to subtract it.
    
    lightBackground - Whether the image has a light background.
    
    useParaboloid - Whether to use the "sliding paraboloid" algorithm.
    
    doPresmooth - Whether the image should be smoothened (3x3 mean) before
    creating the background. With smoothing, the background will not necessarily
    be below the image data.
    
    correctCorners - Whether the algorithm should try to detect corner particles
    to avoid subtracting them as a background.
    '''
    def __init__(self, java_options=None, _imagej_version=None) -> None:
        global IJ
        if ('IJ' not in globals()) or IJ is None:
            if java_options is None:
                java_ram_fraction = 0.7
                java_ram = math.floor(
                    psutil.virtual_memory().available / 1024**2 * java_ram_fraction
                )
                java_options = f"-Xmx{java_ram}m"
            scyjava.config.add_option(java_options)
            print('Java option:', java_options)
            IJ = imagej.init(_imagej_version)
        print('ImageJ Version:', IJ.getVersion())
        self.background_subtracter = jpype.JClass(
            'ij.plugin.filter.BackgroundSubtracter'
        )()
    
    def rolling_ball_background(
        self,
        img,
        radius: float,
        create_background: bool = False,
        light_background: bool = False,
        use_paraboloid: bool = False,
        do_presmooth: bool = False,
        correct_corners: bool = True,
        inplace = False,
    ):
        assert img.ndim == 2
        if not inplace:
            img = np.array(img)
        imp = IJ.py.to_imageplus(img)
        self.background_subtracter.rollingBallBackground(
            imp.getProcessor(),
            radius,
            create_background,
            light_background,
            use_paraboloid,
            do_presmooth,
            correct_corners
        )
        IJ.py.sync_image(imp)
        imp.close()
        imp = None
        return img

    def rolling_ball_background_chunked(
        self, img, radius, chunk_size, compute=False, **kwargs
    ):
        assert img.ndim == 2
        import dask.array as da
        import zarr
        h, w = img.shape
        overlap_depth = radius * get_shrink_factor(radius)    
        chunk_h, overlap_h = compute_chunk_size_and_overlap(chunk_size, overlap_depth, h)
        chunk_w, overlap_w = compute_chunk_size_and_overlap(chunk_size, overlap_depth, w)

        if issubclass(type(img), np.ndarray):
            zimg = zarr.create(
                img.shape,
                chunks=(chunk_h, chunk_w),
                dtype=img.dtype
            )
            zimg[:] = img
            img = None
            img = da.from_zarr(zimg)
        
        if not np.all(img.chunksize == (chunk_h, chunk_w)):
            img = img.rechunk((chunk_h, chunk_w))
        processed = da.map_overlap(
            self.rolling_ball_background, img, dtype=img.dtype, 
            depth={0: overlap_h, 1: overlap_w}, boundary='none', radius=radius,
            **kwargs
        )
        if compute:
            return da_to_zarr(processed)
        return processed


def compute_chunk_size_and_overlap(chunk_size, overlap_depth, img_size):
    # small image
    if (img_size <= chunk_size) or (img_size <= overlap_depth):
        return img_size, 0
    # small chunk
    if chunk_size <= overlap_depth:
        chunk_size = overlap_depth
    chunk_size = math.ceil(chunk_size / 8) * 8
    return chunk_size, overlap_depth


# https://github.com/imagej/imagej1/blob/master/ij/plugin/filter/BackgroundSubtracter.java#L779-L795
def get_shrink_factor(radius):
    if radius <= 10:
        shrinkFactor = 2
    elif radius <= 30:
        shrinkFactor = 2
    elif radius <= 100:
        shrinkFactor = 4
    else:
        shrinkFactor = 8
    return shrinkFactor


def da_to_zarr(da_img, zarr_store=None, num_workers=None, out_shape=None, chunks=None):
    import zarr
    if zarr_store is None:
        if out_shape is None:
            out_shape = da_img.shape
        if chunks is None:
            chunks = da_img.chunksize
        zarr_store = zarr.create(
            out_shape,
            chunks=chunks,
            dtype=da_img.dtype,
            overwrite=True
        )
    da_img.to_zarr(zarr_store, compute=False).compute(
        num_workers=num_workers
    )
    return zarr_store



import pathlib
import sys

import dask.array as da
import dask.config
import fire
import palom
import tifffile
from loguru import logger
from palom.cli import pyramid_tools as ptools

from .. import BackgroundSubtracter
from .. import __version__ as _version

ROLLING_BALL_DEFAULT = {
    "create_background": False,
    "light_background": False,
    "use_paraboloid": False,
    "do_presmooth": False,
    "correct_corners": True,
}


def process_ometiff(
    img_path,
    radius,
    _channel: int = None,
    out_path: str = None,
    target_chunk_size: int = 1024 * 5,
    overwrite: bool = False,
    java_options: str = None,
    imagej_version: str = None,
    pyramid_config: dict = None,
    rolling_ball_kwargs: dict = None,
    num_workers: int = 4,
):
    img_path = pathlib.Path(img_path).absolute()
    reader = palom.reader.OmePyramidReader(img_path)
    stem = img_path.name.split(".")[0]
    out_path = ptools.validate_out_path(
        # FIXME out_path should be full path; this should be handled by
        # `validate_out_path`
        out_path if out_path is None else pathlib.Path(out_path).absolute(),
        img_path.parent / f"{stem}-ij_rolling_ball_{radius}.ome.tif",
        overwrite=overwrite,
    )

    log_str = f"""

    Processing: {img_path.name}
    Rolling ball radius: {radius}
    Input shape: {reader.pyramid[0].shape}
    Output path: {out_path}

    """
    logger.info(log_str)
    if java_options is not None:
        logger.info(f"Java option: {java_options}")

    img = reader.pyramid[0]
    if _channel is not None:
        img = img[_channel : _channel + 1]
    bg_subtracter = BackgroundSubtracter(
        java_options=java_options, _imagej_version=imagej_version
    )

    if rolling_ball_kwargs is None:
        rolling_ball_kwargs = {}
    out = []
    for channel in img:
        out.append(
            bg_subtracter.rolling_ball_background_chunked(
                channel,
                radius,
                target_chunk_size,
                **{**ROLLING_BALL_DEFAULT, **rolling_ball_kwargs},
            )
        )
    try:
        tif_tags = ptools.src_tif_tags(img_path)
    except Exception:
        tif_tags = {}
    tif_tags.update({"software": f"imagej-rolling-ball v{_version}"})
    # FIXME handle channel names?
    if pyramid_config is None:
        pyramid_config = {}

    with dask.config.set(scheduler="threads", num_workers=num_workers):
        palom.pyramid.write_pyramid(
            [da.array(out)],
            output_path=out_path,
            **{
                **dict(
                    pixel_size=reader.pixel_size,
                    kwargs_tifffile=tif_tags,
                ),
                **dict(
                    downscale_factor=2,
                    compression="zlib",
                    tile_size=1024,
                    save_RAM=True,
                ),
                **pyramid_config,
            },
        )
    try:
        ome_xml = tifffile.tiffcomment(img_path)
        tifffile.tiffcomment(out_path, ome_xml.encode())
    except Exception:
        logger.warning("Unable to transfer original ome-xml from input image")
    return 0


def main():
    sys.exit(fire.Fire(process_ometiff))


if __name__ == "__main__":
    main()
